from .logging import grown_log
from .store import storage as _storage
from userv.routing import Router as _Router
from userv.async_server import run_server as _runserver
from .wlan import connect_and_configure_wlan as _connect_and_configure_wlan
from .time_control import time_sync_task as _time_sync_task

try:
    import uasyncio as asyncio
except ImportError:
    import asyncio


def setup():
    """
    setting up grown tasks server

    This setup provides basic functions like connection to wlan or setting up a regular hotspot.

    configurable by a rest interface
    further functions can be added by each control

    :rtype: userv.routing.Router
    """

    # create tasks
    loop = asyncio.get_event_loop()
    # setup time sync
    loop.create_task(_time_sync_task())

    # adding routes
    # routes can be added till a run is triggered
    router = _Router()

    # configurable tasks
    # wlan
    _connect_and_configure_wlan(router)

    # TODO add also swagger ui

    # adding async webserver
    _runserver(router)
    return router


def run_grown():
    """
    runs a grown application
    """
    try:
        asyncio.get_event_loop().run_forever()
    except Exception as e:
        grown_log.error(str(e))
    finally:
        # TODO save log to file?
        pass



# import gc
# import os
import time
from .store import storage
from userv.routing import json_response
from .logging import grown_log

try:
    import uasyncio as asyncio
except ImportError:
    import asyncio


# TODO history of data
# _history_sensor_data_file = "sensor_data_log.json"


# async def _trim_sensor_data(total_lines, keep_lines):
#     os.rename(_history_sensor_data_file, "temp.json")
#     file_ptr = open("temp.json", "r")
#     for _ in range(0, total_lines - keep_lines):
#         file_ptr.readline()
#         gc.collect()
#     new_file_ptr = open(_history_sensor_data_file, "a")
#     while True:
#         read = file_ptr.readline()
#         if read == "":
#             break
#         else:
#             new_file_ptr.write(read)
#         gc.collect()
#
#     new_file_ptr.close()
#     file_ptr.close()
#     os.remove("temp.json")
#
#
# async def _save_sensor_data(sensor_data):
#     string_to_write = json.dumps(sensor_data) + "\n"
#     file_ptr = open(_history_sensor_data_file, "a")
#     file_ptr.write(json.dumps(sensor_data) + "\n")
#     end_byte = file_ptr.tell()
#     file_ptr.close()
#     assumed_total_lines = int(end_byte) // len(string_to_write)
#     if assumed_total_lines > 1200:
#         await _trim_sensor_data(assumed_total_lines, 1000)


def _update_reducer(store_dict, data):
    """
    :type store_dict: dict
    :type data: dict
    :rtype: dict
    """
    new_sensor_data = dict(time=time.time())
    new_sensor_data.update(data)
    return new_sensor_data


async def _sensor_data_task(get_sensor_data):
    """
    task for continues logging
    """
    sensor_data_leaf = storage.get_leaf('sensor_data')

    while True:
        try:
            if isinstance(get_sensor_data, type(lambda: (yield))):
                data = await get_sensor_data()
            else:
                data = get_sensor_data()
            sensor_data_leaf.update(data)
        except Exception as e:
            grown_log.error(str(e))
        # TODO sleep from settings
        await asyncio.sleep(60)


async def _get_sensor_data(request):
    """
    path for actual data
    """
    data_leaf = storage.get_leaf('sensor_data')
    return json_response(data_leaf.get())


def add_data_control(router, gather_data_func):
    """
    :type router: userv.routing.Router
    :param gather_data_func: gather data function
    """
    grown_log.info("Adding data control to grown.")
    try:
        assert callable(gather_data_func) is True, "gather_data_func is not callable"
        # create tasks
        loop = asyncio.get_event_loop()

        storage.register_leaf('sensor_data', {'time': time.time()}, _update_reducer)
        loop.create_task(_sensor_data_task(gather_data_func))

        # adding routes
        router.add("/rest/sensor_data", _get_sensor_data)
    except Exception as e:
        grown_log.error(str(e))
    # TODO history data



"""
rudimentary monkey_patch till module has fileio write
"""
import ulogging


def log(self, level, msg, *args):
    _log_file = open('run_information.log', 'a')
    _log_file.write("%s:%s: %s \n" % (self._level_str(level), self.name, msg))
    _log_file.close()
    if level >= (self.level or ulogging._level):
        ulogging._stream.write("%s:%s:" % (self._level_str(level), self.name))
        if not args:
            print(msg, file=ulogging._stream)
        else:
            print(msg % args, file=ulogging._stream)


ulogging.Logger.log = log


ulogging.basicConfig(
    ulogging.DEBUG,
)
grown_log = ulogging.getLogger("grown")



"""
module for time
"""
import machine
import utime
from .logging import grown_log

try:
    import uasyncio as asyncio
except ImportError:
    import asyncio
try:
    import usocket as socket
except ImportError:
    import socket
try:
    import ustruct as struct
except ImportError:
    import struct

# (date(2000, 1, 1) - date(1900, 1, 1)).days * 24*60*60
NTP_DELTA = 3155673600

host = "pool.ntp.org"


def _set_system_time(ntp_time):
    """
    There's currently no timezone support in MicroPython, so
    utime.localtime() will return UTC time (as if it was .gmtime())
    """
    tm = utime.localtime(ntp_time)
    tm = tm[0:3] + (0,) + tm[3:6] + (0,)
    machine.RTC().datetime(tm)
    grown_log.info("timesync: %s" % str(utime.localtime()))


async def _time_from_server():
    NTP_QUERY = bytearray(48)
    NTP_QUERY[0] = 0x1b
    addr = socket.getaddrinfo(host, 123)[0][-1]
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(1)
    res = s.sendto(NTP_QUERY, addr)
    msg = s.recv(48)
    s.close()
    val = struct.unpack("!I", msg[40:44])[0]
    return val - NTP_DELTA


async def time_sync_task(time_between_syncs_s=300):
    """
    task for synchronising time
    """
    while True:
        try:
            server_time = await _time_from_server()
            _set_system_time(server_time)
        except Exception as e:
            grown_log.error(str(e))

        await asyncio.sleep(time_between_syncs_s)


def get_current_time():
    """
    retuns the current hh:mm:ss in seconds
    :rtype: int
    """
    current_time = utime.time()
    return current_time % (60*60*24)



from .store import storage
from userv.routing import json_response, text_response
from .logging import grown_log
from grown.time_control import get_current_time
try:
    import ujson as json
except ImportError:
    import json
try:
    import uasyncio as asyncio
except ImportError:
    import asyncio


async def _light_control_task(enable_func, disable_func, safety_function):
    """
    all parameter are async functions and running in an infinite loop.
    control parameter

    :param enable_func: no parameter just enable light
    :param disable_func: no parameter just enable light
    :param safety_function: safety functions is always called before. returned true if not safe to run
    """
    if safety_function is None:
        safety_function = lambda x: False
    data_leaf = storage.get_leaf('light_control')
    sensor_data_leaf = storage.get_leaf('sensor_data')
    last_state = None
    while data_leaf is not None:
        try:
            data = data_leaf.get()
            current_time = get_current_time()
            # only need time by day
            sensor_data = sensor_data_leaf.get()

            result = safety_function(sensor_data)
            if isinstance(safety_function, type(lambda: (yield))):
                result = await result
            if result is not True:
                switch_on_time = data['switch_on_time']
                switch_off_time = data['switch_off_time']
                # if switching off lies in the next day
                if switch_on_time > switch_off_time:
                    switch_off_time += 24 * 60 * 60

                if switch_on_time <= current_time <= switch_off_time:
                    # Light on
                    if last_state is not True:
                        grown_log.info("light_control: enable light")
                        last_state = True
                    enable = enable_func()
                    if isinstance(enable_func, type(lambda: (yield))):
                        await enable
                else:
                    # Light off
                    if last_state is not False:
                        grown_log.info("light_control: disable light")
                        last_state = False
                    disable = disable_func()
                    if isinstance(disable_func, type(lambda: (yield))):
                        await disable
        except Exception as e:
            grown_log.error(str(e))
        await asyncio.sleep(100)


async def _get_light_control_data(request):
    """
    path for actual data
    """
    data_leaf = storage.get_leaf('light_control')
    return json_response(data_leaf.get())


async def _post_light_control_data(request):
    leaf = storage.get_leaf('light_control')
    try:
        new_light_control = json.loads(request.get('body', ""))
        leaf.update(new_light_control)
        return json_response(new_light_control)
    except Exception as e:
        return text_response(str(e), status=400)


def _update_reducer(store_dict, data):
    """
    :type store_dict: dict
    :type data: dict
    :rtype: dict
    """
    store_dict.update(data)
    return store_dict


def add_light_control(router, enable_func, disable_func, safety_function=None):
    """
    Adds an element controling light to the plants. Further it sets up a task regulating the light which can be
    configured via rest allowing other tasks from outside to optimise the task at hand.

    :type router: user.routing.Router
    :param enable_func: async function to enable light
    :param disable_func: async function to disable light
    :param safety_function:
    """
    grown_log.info('Adding light control to grown server')
    try:
        assert callable(enable_func) is True, "enable_func is not callable"
        assert callable(disable_func) is True, "enable_func is not callable"
        storage.register_leaf(
            'light_control',
            {
                'switch_on_time': 11 * 3600,
                'switch_off_time': 23 * 3600
            },
            _update_reducer
        )
        # create lighting task based on set settings
        loop = asyncio.get_event_loop()
        loop.create_task(_light_control_task(enable_func, disable_func, safety_function))
        # create subserver for light control
        router.add("/rest/light_control", _get_light_control_data, 'GET')
        router.add("/rest/light_control", _post_light_control_data, 'POST')
    except Exception as e:
        grown_log.error(str(e))




"""
Module to control all settings data and so on
Usage should be 1 instance of store storing and transmitting to elements
The way python works on load creates a single instance and modules are normaly not loaded twice
"""
try:
    import ujson as json
except ImportError:
    import json
import os
_settings_file = "settings_store.json"


class _Leaf():

    def __init__(self, initial_data=None, update_reducer=None, save_callback=None):
        if initial_data is None:
            initial_data = {}
        self._store = initial_data
        self._update_reducer = update_reducer
        self._save_callback = save_callback

    def update(self, data):
        """
        updates the data therefore the update_reducer is invoked and data dumped to flash drive
        """
        if self._update_reducer is None:
            return
        self._store = self._update_reducer(self._store, data)
        if self._save_callback is None:
            return
        self._save_callback()

    def get(self, key=None):
        """
        returns the contents of the leaf if a get is given only this subset is returned
        """
        if key is None:
            return self._store
        return self._store.get(key, None)


class _Store():

    def __init__(self):
        if _settings_file not in os.listdir():
            self._store = {}
        else:
            file_ptr = open(_settings_file, "r")
            try:
                self._store = json.loads(file_ptr.read())
            except ValueError:
                os.remove(_settings_file)
                # TODO log error
            finally:
                file_ptr.close()

    def _dump_data(self):
        """
        pickeles hole object to one dict
        :rtype: dict
        """
        return {key: value if type(value) == dict else value.get() for key, value in self._store.items()}

    def save(self):
        """stores all leafs to an json pickeled file"""
        # save config

        file_ptr = open(_settings_file, "w")
        try:
            file_ptr.write(json.dumps(self._dump_data()))
        finally:
            file_ptr.close()

    def register_leaf(self, key, initial_data, update_function):
        """
        registeres a new lead to store data
        :rtype initial_data: dict
        :param update_function: function with 2 parameter old_data, new_data
        """
        initial_data = self._store.get(key, initial_data)
        if isinstance(initial_data, _Leaf):
            return initial_data
        self._store[key] = _Leaf(
            initial_data,
            update_reducer=update_function,
            save_callback=self.save
        )

    def get_leaf(self, key):
        """
        retruns the current state of a leaf
        :rtype: _Leaf
        """
        leaf = self._store.get(key, {})
        if isinstance(leaf, dict):
            return _Leaf(leaf)
        return leaf


storage = _Store()



import network
import time
from .store import storage
from .logging import grown_log
from userv.routing import json_response, text_response
try:
    import ujson as json
except ImportError:
    import json

_sta_if = network.WLAN(network.STA_IF)
_ap_if = network.WLAN(network.AP_IF)


def _connect_to_existing_network(essid, password):
    if essid is None or password is None:
        grown_log.info('essid or password is not set')
        return False

    grown_log.info('connecting to network...')
    _sta_if.active(True)

    _sta_if.connect(essid, password)
    start_time = time.time()
    grown_log.info("connecting")
    while not _sta_if.isconnected() and start_time + 10 >= time.time():
        print(".")
        time.sleep(1)
    if not _sta_if.isconnected():
        _sta_if.active(False)
        grown_log.info('Connection NOT establisched')
        return False
    else:
        grown_log.info('network config: %s' % str(_sta_if.ifconfig()))
    return True


def _create_an_network():
    _ap_if.active(True)
    essid = "MyPlantMonitor"
    password = "MyPlantMonitor"
    _ap_if.config(essid=essid, password=password, authmode=4)
    grown_log.info('essid: %s  pw: %s' % (essid, password))


def _update_wlan_data(old_data, new_data):
    """
    :type old_data: dict
    :type new_data: dict
    :rtype: dict
    """
    old_data.update(new_data)
    return old_data


def _get_wlan_config(request):
    # we should not transmit the password
    wlan_config = storage.get_leaf('wlan')
    wlan_config['password'] = "*" * len(wlan_config['password'])
    return json_response(wlan_config)


def _post_wlan_config(request):
    leaf = storage.get_leaf('wlan')
    try:
        new_wlan_config = json.loads(request.get('body', ""))
        leaf.update(new_wlan_config)
        return json_response(new_wlan_config)
    except Exception as e:
        return text_response(str(e), status=400)


def connect_and_configure_wlan(router):
    """
    configures and starts wlan server.
    Also adds routes to server
    :type router: userv.routing.Router
    """
    # connect
    storage.register_leaf(
        'wlan',
        dict(
            ssid=None,
            password=None,
        ),
        _update_wlan_data
    )

    wlan_config = storage.get_leaf('wlan')

    grown_log.info("connect existing network")
    network_connected = _connect_to_existing_network(
        essid=wlan_config.get('ssid'),
        password=wlan_config.get('password'),
    )
    if not network_connected:
        grown_log.info("Create hotspot")
        _create_an_network()

    #
    router.add("/setting/wlan", _get_wlan_config, method="GET")
    router.add("/setting/wlan", _post_wlan_config, method="POST")
